@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 
 import warnings
+warnings.simplefilter("once", UserWarning)
 import numpy as np
 from loguru import logger
 from distributed import Client
@@ -11,17 +12,20 @@ from .gp_marginal_density import GPMarginalDensity
 from .gp_likelihood import GPlikelihood
 from .gp_training import GPtraining
 from .gp_posterior import GPposterior
+import importlib
 
 
 # TODO: search below "TODO"
-#   as work on functions is completed and functions are maintained, add verbose assert statements.
+#   - as work on functions is completed and functions are maintained, add verbose assert statements.
+#   - make shape(y_data)=(V,U) work (tensor GP), change docstrings
 
 
 class GP:
     """
     This class provides all the tools for a single-task Gaussian Process (GP).
     Use fvGP for multi-task GPs. However, the fvGP class inherits all methods from this class.
-    This class allows full HPC support for training via the `hgdl` package.
+    This class allows full HPC distributed training via the `hgdl` package, large-scale sparse GPs
+    via `gp2Scale` and offers GPU support.
 
     V ... number of input points
 
@@ -38,12 +42,13 @@ class GP:
         For multi-task GPs, the index set dimension = input space dimension + 1.
         If dealing with non-Euclidean inputs
         x_data should be a list, not a numpy array.
+        In this case, both the index set and the input space dim are set to 1.
     y_data : np.ndarray
         The values of the data points. Shape (V).
     init_hyperparameters : np.ndarray, optional
         Vector of hyperparameters used to initiate the GP.
         The default is an array of ones with the right length for the anisotropic Matern
-        kernel with automatic relevance determination (ARD). If gp2Scale is
+        kernel with automatic relevance determination (ARD). If `gp2Scale` is
         enabled, the default kernel changes to the anisotropic Wendland kernel.
     noise_variances : np.ndarray, optional
         An numpy array defining the uncertainties/noise in the
@@ -55,8 +60,8 @@ class GP:
         Only provide a noise function OR `noise_variances`, not both.
     compute_device : str, optional
         One of `cpu` or `gpu`, determines how linear algebra computations are executed. The default is `cpu`.
-        For "gpu", pytorch has to be installed manually.
-        If gp2Scale is enabled but no kernel is provided, the choice of the `compute_device`
+        For `gpu`, pytorch or cupy has to be installed manually. For advanced options see `args`.
+        If `gp2Scale` is enabled but no kernel is provided, the choice of the `compute_device`
         will be particularly important. In that case, the default Wendland kernel will be computed on
         the cpu or the gpu which will significantly change the compute time depending on the compute
         architecture.
@@ -78,7 +83,7 @@ class GP:
         `x2` (a N2 x D array of positions) and
         `hyperparameters` (a 1d array of length D+1 for the default kernel).
         The default is a finite difference calculation.
-        If `ram_economy` is True, the function's input is x1, x2, direction (int), and hyperparameters (numpy array).
+        If `ram_economy` is True, the function's input is x1, x2, hyperparameters (numpy array), and a direction (int).
         The output is a numpy array of shape (len(hps) x N).
         If `ram_economy` is `False`, the function's input is x1, x2, and hyperparameters.
         The output is a numpy array of shape (len(hyperparameters) x N1 x N2). See `ram_economy`.
@@ -116,6 +121,7 @@ class GP:
         hyperparameters, or, if `noise_function` is provided but no noise function,
         a finite-difference approximation will be used.
         The same rules regarding `ram_economy` as for the kernel definition apply here.
+        That means the function will have an additional `direction` parameter.
     gp2Scale: bool, optional
         Turns on gp2Scale. This will distribute the covariance computations across multiple workers.
         This is an advanced feature for HPC GPs up to 10
@@ -139,21 +145,16 @@ class GP:
         to compute a factorization object
         which is available in the second and third callable. The second being the linear solve f(obj, vec),
         and the third being the logdet=f(obj). If a factorization object is not required, the first callable
-        can return the matrix itself (K).
+        should return the matrix itself (K).
     calc_inv : bool, optional
         If True, the algorithm calculates and stores the inverse of the covariance
         matrix after each training or update of the dataset or hyperparameters,
         which makes computing the posterior covariance faster (3-10 times).
-        For larger problems (>2000 data points), the use of inversion should be avoided due
+        For larger problems (>5000 data points), the use of inversion should be avoided due
         to computational instability and costs. The default is
         False. Note, the training will not use the
         inverse for stability reasons. Storing the inverse is
         a good option when the dataset is not too large and the posterior covariance is heavily used.
-        Caution: this option, together with `append=True` in `tell()` will mean that the inverse of
-        the covariance is updated, not recomputed, which can lead to instability.
-        In application where data is appended many times, it is recommended to either turn
-        `calc_inv` off, or to regularly force the recomputation of the inverse via `gp_rank_n_update` in
-        `update_gp_data`.
     ram_economy : bool, optional
         Only of interest if the gradient and/or Hessian of the log marginal likelihood is/are used for the training.
         If True, components of the derivative of the log marginal likelihood are
@@ -161,32 +162,47 @@ class GP:
         but much less RAM usage. If the derivative of the kernel (and noise function) with
         respect to the hyperparameters (kernel_function_grad) is
         going to be provided, it has to be tailored: for `ram_economy=True` it should be
-        of the form f(x, direction, hyperparameters)
+        of the form f(x, hyperparameters, direction)
         and return a 2d numpy array of shape len(x1) x len(x2).
         If `ram_economy=False`, the function should be of the form f(x, hyperparameters)
         and return a numpy array of shape
         H x len(x1) x len(x2), where H is the number of hyperparameters.
         CAUTION: This array will be stored and is very large.
     args: dict, optional
-        A dictionary of advances settings.
+        Advanced options. Recognized keys are:
+
+        - "random_logdet_lanczos_degree" : int; default = 20
+        - "random_logdet_error_rtol" : float; default = 0.01
+        - "random_logdet_verbose" : True/False; default = False
+        - "random_logdet_print_info" : True/False; default = False
+        - "sparse_minres_tol" : float
+        - "cg_minres_tol" : float
+        - "random_logdet_lanczos_compute_device" : str; default = "cpu"/"gpu"
+        - "Chol_factor_compute_device" : str; default = "cpu"/"gpu"
+        - "update_Chol_factor_compute_device": str; default = "cpu"/"gpu"
+        - "Chol_solve_compute_device" : str; default = "cpu"/"gpu"
+        - "Chol_logdet_compute_device" : str; default = "cpu"/"gpu"
+        - "GPU_engine" : str; default = "torch"/"cupy"
+
+        All other keys will be stored and are available as part of the object instance.
 
 
     Attributes
     ----------
     x_data : np.ndarray
-        Datapoint positions
+        Datapoint positions.
     y_data : np.ndarray
-        Datapoint values
+        Datapoint values.
     noise_variances : np.ndarray
-        Datapoint observation variances
-    prior.hyperparameters : np.ndarray
+        Datapoint observation variances.
+    hyperparameters : np.ndarray
         Current hyperparameters in use.
-    prior.K : np.ndarray
-        Current prior covariance matrix of the GP
-    prior.m : np.ndarray
+    K : np.ndarray
+        Current prior covariance matrix of the GP.
+    m : np.ndarray
         Current prior mean vector.
-    likelihood.V : np.ndarray
-        the noise covariance matrix
+    V : np.ndarray
+        the noise covariance matrix.
     """
 
     def __init__(
@@ -210,8 +226,7 @@ class GP:
         ram_economy=False,
         args=None
     ):
-        assert isinstance(x_data, list) or isinstance(x_data, np.ndarray), "wrong format in x_data"
-        assert isinstance(y_data, np.ndarray) and np.ndim(y_data) == 1, "wrong format in y_data"
+
         assert isinstance(noise_variances, np.ndarray) or noise_variances is None, "wrong format in noise_variances"
         assert init_hyperparameters is None or isinstance(init_hyperparameters,np.ndarray), "wrong init_hyperparameters"
         assert isinstance(compute_device, str), "wrong format in compute_device"
@@ -224,17 +239,19 @@ class GP:
             "wrong format in prior_mean_function"
         assert len(x_data) == len(y_data), "x_data and y_data do not have the same lengths."
 
-        self.compute_device = compute_device
-        self.calc_inv = calc_inv
-        self.gp2Scale = gp2Scale
-        if args is None: self.args = {}
-        else: self.args = args
+        if args is None: args = {}
         hyperparameters = init_hyperparameters
 
         ########################################
         ###init data instance###################
         ########################################
-        self.data = GPdata(x_data, y_data, noise_variances)
+        self.data = GPdata(x_data, y_data,
+                           args=args,
+                           noise_variances=noise_variances,
+                           ram_economy=ram_economy,
+                           gp2Scale=gp2Scale,
+                           compute_device=compute_device,
+                           calc_inv=calc_inv)
         ########################################
         # prepare initial hyperparameters and bounds
         if self.data.Euclidean:
@@ -244,7 +261,7 @@ class GP:
                     "initial hyperparameters.")
             else:
                 if init_hyperparameters is None:
-                    hyperparameters = np.ones((self.data.index_set_dim + 1))
+                    hyperparameters = np.ones((self.index_set_dim + 1))
                     warnings.warn("Hyperparameters initialized to a vector of ones.")
         else:
             hyperparameters = init_hyperparameters
@@ -253,49 +270,36 @@ class GP:
         if hyperparameters is None:
             raise Exception("'init_hyperparameters' not provided and could not be calculated. Please provide them ")
 
-        if gp2Scale:
-            try:
-                from imate import logdet as imate_logdet
-            except:
-                raise Exception(
-                    "You have activated `gp2Scale`. You need to install imate"
-                    " manually for this to work.")
-            if gp2Scale_dask_client is None:
-                logger.debug("Creating my own local client.")
-                gp2Scale_dask_client = Client()
-
         if compute_device == 'gpu':
-            try:
-                import torch
-            except:
-                raise Exception(
-                    "You have specified the 'gpu' as your compute device. You need to install pytorch "
-                    "manually for this to work.")
+            if not importlib.util.find_spec("torch") and not importlib.util.find_spec("cupy"):
+                warnings.warn("You have specified the 'gpu' as your compute device. You need to install pytorch or cupy"
+                              " manually for this to work.")
+        # Check gp2Scale
+        gp2Scale_dask_client = self.initialize_gp2Scale_dask_client(gp2Scale, gp2Scale_dask_client)
+
+        ##########################################
+        #######prepare training###################
+        ##########################################
+        self.trainer = GPtraining(self.data, hyperparameters)
         ########################################
         ###init prior instance##################
         ########################################
         self.prior = GPprior(self.data,
-                             hyperparameters=hyperparameters,
+                             self.trainer,
                              kernel=kernel_function,
                              prior_mean_function=prior_mean_function,
                              kernel_grad=kernel_function_grad,
                              prior_mean_function_grad=prior_mean_function_grad,
-                             gp2Scale=gp2Scale,
                              gp2Scale_dask_client=gp2Scale_dask_client,
                              gp2Scale_batch_size=gp2Scale_batch_size,
-                             ram_economy=ram_economy,
-                             args=self.args
                              )
         ########################################
         ###init likelihood instance#############
         ########################################
         self.likelihood = GPlikelihood(self.data,
-                                       hyperparameters=self.prior.hyperparameters,
+                                       self.trainer,
                                        noise_function=noise_function,
                                        noise_function_grad=noise_function_grad,
-                                       ram_economy=ram_economy,
-                                       gp2Scale=gp2Scale,
-                                       args=self.args
                                        )
 
         ##########################################
@@ -305,31 +309,100 @@ class GP:
             self.data,
             self.prior,
             self.likelihood,
-            calc_inv=calc_inv,
-            gp2Scale=gp2Scale,
+            self.trainer,
             gp2Scale_linalg_mode=gp2Scale_linalg_mode,
-            compute_device=compute_device,
-            args=self.args
         )
 
-        ##########################################
-        #######prepare training###################
-        ##########################################
-        self.trainer = GPtraining(gp2Scale=gp2Scale, args=self.args)
-
         ##########################################
         #######prepare posterior evaluations######
         ##########################################
         self.posterior = GPposterior(self.data,
                                      self.prior,
+                                     self.trainer,
                                      self.marginal_density,
-                                     self.likelihood,
-                                     args=self.args
-                                     )
-        self.x_data = self.data.x_data
-        self.y_data = self.data.y_data
-        self.noise_variances = self.data.noise_variances
-        self.index_set_dim = self.data.index_set_dim
+                                     self.likelihood)
+
+    #########PROPERTIES#########################################
+    @property
+    def x_data(self):
+        return self.data.x_data
+
+    @property
+    def y_data(self):
+        return self.data.y_data
+
+    @property
+    def noise_variances(self):
+        return self.data.noise_variances
+
+    @property
+    def index_set_dim(self):
+        return self.data.index_set_dim
+
+    @property
+    def input_set_dim(self):
+        return self.data.input_set_dim
+
+    @property
+    def mcmc_info(self):
+        return self.trainer.mcmc_info
+
+    @property
+    def args(self):
+        return self.data.args
+
+    @args.setter
+    def args(self, args):
+        self.data.args = args
+
+    @property
+    def K(self):
+        return self.prior.K
+
+    @property
+    def m(self):
+        return self.prior.m
+
+    @property
+    def V(self):
+        return self.likelihood.V
+
+    @property
+    def hyperparameters(self):
+        return self.trainer.hyperparameters
+
+    @property
+    def gp2Scale(self):
+        return self.data.gp2Scale
+    ###############################################################
+    def set_args(self, new_args):
+        """
+        Use this function to change the arguments for the GP.
+
+        Parameters
+        ----------
+        new_args : dict
+            The new advanced settings.
+        """
+        self.args = new_args
+
+    ##################################################################################
+    def set_hyperparameters(self, hps):
+        """
+        Function to set hyperparameters.
+
+
+        Parameters
+        ----------
+        hps : np.ndarray
+            A 1-d numpy array of hyperparameters.
+        """
+        assert isinstance(hps, np.ndarray), "wrong format in hyperparameters"
+        assert np.ndim(hps) == 1, "wrong format in hyperparameters"
+        self.trainer.hyperparameters = hps
+        self.prior.update_state_hyperparameters()
+        self.likelihood.update_state()
+        self.marginal_density.update_state_hyperparameters()
 
     def update_gp_data(
         self,
@@ -356,7 +429,7 @@ class GP:
             The values of the data points. Shape (V).
         noise_variances_new : np.ndarray, optional
             An numpy array defining the uncertainties in the data `y_data` in form of a point-wise variance.
-            Shape (len(y_data)).
+            Shape(y_new)==Shape(noise_variances_new).
             Note: if no variances are provided here, the noise_covariance
             callable will be used; if the callable is not provided the noise variances
             will be set to `abs(np.mean(y_data)) / 100.0`. If you provided a noise function at initialization,
@@ -370,40 +443,25 @@ class GP:
             be performed.
         """
         assert isinstance(x_new, list) or isinstance(x_new, np.ndarray), "wrong format in x_new"
-        assert isinstance(y_new, np.ndarray) and np.ndim(y_new) == 1, "wrong format in y_new"
-        assert isinstance(noise_variances_new, np.ndarray) or noise_variances_new is None, "wrong format in noise_variances_new"
+        assert isinstance(y_new, np.ndarray) and (np.ndim(y_new) == 1 or np.ndim(y_new) == 2), "wrong format in y_new"
+        assert isinstance(noise_variances_new, np.ndarray) or noise_variances_new is None, \
+            "wrong format in noise_variances_new"
         assert len(x_new) == len(y_new), "updated x and y do not have the same lengths."
-        old_x_data = self.data.x_data.copy()
+        old_x_data = self.x_data.copy()
         if gp_rank_n_update is None: gp_rank_n_update = append
         # update data
         self.data.update(x_new, y_new, noise_variances_new, append=append)
 
         # update prior
-        if append:
-            self.prior.augment_data(old_x_data, x_new)
-        else:
-            self.prior.update_data()
+        if append: self.prior.augment_state_data(old_x_data, x_new)
+        else: self.prior.update_state_data()
 
         # update likelihood
-        self.likelihood.update(self.prior.hyperparameters)
+        self.likelihood.update_state()
 
         # update marginal density
-        self.marginal_density.update_data(gp_rank_n_update)
+        self.marginal_density.update_state_data(gp_rank_n_update)
         ##########################################
-        self.x_data = self.data.x_data
-        self.y_data = self.data.y_data
-
-    def set_args(self, args):
-        self.args = args
-        self.prior.args = args
-        self.likelihood.args = args
-        self.marginal_density.args = args
-        self.trainer.args = args
-        self.posterior.args = args
-        self.marginal_density.KVlinalg.args = args
-
-    def get_args(self):
-        return self.args
 
     def _get_default_hyperparameter_bounds(self):
         """
@@ -417,13 +475,13 @@ class GP:
         """
         if not self.data.Euclidean: raise Exception("Please provide custom hyperparameter bounds to "
                                                     "the training in the non-Euclidean setting")
-        if len(self.prior.hyperparameters) != self.data.index_set_dim + 1:
+        if len(self.hyperparameters) != self.index_set_dim + 1:
             raise Exception("Please provide custom hyperparameter_bounds when kernel, mean or noise"
                             " functions are customized")
-        hyperparameter_bounds = np.zeros((self.data.index_set_dim + 1, 2))
-        hyperparameter_bounds[0] = np.array([np.var(self.data.y_data) / 100., np.var(self.data.y_data) * 10.])
-        for i in range(self.data.index_set_dim):
-            range_xi = np.max(self.data.x_data[:, i]) - np.min(self.data.x_data[:, i])
+        hyperparameter_bounds = np.zeros((self.index_set_dim + 1, 2))
+        hyperparameter_bounds[0] = np.array([np.var(self.y_data) / 100., np.var(self.y_data) * 10.])
+        for i in range(self.index_set_dim):
+            range_xi = np.max(self.x_data[:, i]) - np.min(self.x_data[:, i])
             hyperparameter_bounds[i + 1] = np.array([range_xi / 100., range_xi * 10.])
         assert isinstance(hyperparameter_bounds, np.ndarray) and np.ndim(hyperparameter_bounds) == 2
         return hyperparameter_bounds
@@ -442,7 +500,7 @@ class GP:
               method="mcmc",
               pop_size=20,
               tolerance=0.0001,
-              max_iter=200,
+              max_iter=1000,
               local_optimizer="L-BFGS-B",
               global_optimizer="genetic",
               constraints=(),
@@ -494,7 +552,7 @@ class GP:
         tolerance : float, optional
             Used as termination criterion for local optimizers. Default = 0.0001.
         max_iter : int, optional
-            Maximum number of iterations for global and local optimizers. Default = 120.
+            Maximum number of iterations for global and local optimizers. Default = 1000.
         local_optimizer : str, optional
             Defining the local optimizer. Default = `L-BFGS-B`, most `scipy.optimize.minimize`
             functions are permissible.
@@ -525,12 +583,12 @@ class GP:
                           " mean, or noise functions")
 
         if init_hyperparameters is None:
-            if out_of_bounds(self.prior.hyperparameters, hyperparameter_bounds):
+            if out_of_bounds(self.hyperparameters, hyperparameter_bounds):
                 init_hyperparameters = np.random.uniform(low=hyperparameter_bounds[:, 0],
                                                          high=hyperparameter_bounds[:, 1],
                                                          size=len(hyperparameter_bounds))
             else:
-                init_hyperparameters = self.prior.hyperparameters
+                init_hyperparameters = self.hyperparameters
         else:
             if out_of_bounds(init_hyperparameters, hyperparameter_bounds):
                 warnings.warn("Your init_hyperparameters are out of bounds. They will be over-written")
@@ -569,11 +627,7 @@ class GP:
             dask_client=dask_client,
             info=info
         )
-
-        self.prior.update_hyperparameters(hyperparameters)
-        self.likelihood.update(self.prior.hyperparameters)
-        self.marginal_density.update_hyperparameters()
-        assert isinstance(hyperparameters, np.ndarray) and np.ndim(hyperparameters) == 1
+        self.set_hyperparameters(hyperparameters)
         return hyperparameters
 
     ##################################################################################
@@ -649,12 +703,12 @@ class GP:
                           " mean, or noise functions")
 
         if init_hyperparameters is None:
-            if out_of_bounds(self.prior.hyperparameters, hyperparameter_bounds):
+            if out_of_bounds(self.hyperparameters, hyperparameter_bounds):
                 init_hyperparameters = np.random.uniform(low=hyperparameter_bounds[:, 0],
                                                          high=hyperparameter_bounds[:, 1],
                                                          size=len(hyperparameter_bounds))
             else:
-                init_hyperparameters = self.prior.hyperparameters
+                init_hyperparameters = self.hyperparameters
         else:
             if out_of_bounds(init_hyperparameters, hyperparameter_bounds):
                 warnings.warn("Your init_hyperparameters are out of bounds. They will be over-written")
@@ -723,45 +777,9 @@ class GP:
         hyperparameters : np.ndarray
         """
 
-        res = self.trainer.update_hyperparameters(opt_obj)
-        if res is not None:
-            l_n = self.marginal_density.neg_log_likelihood(res)
-            l_o = self.marginal_density.neg_log_likelihood()
-            if l_n - l_o < 0.000001:
-                hyperparameters = res
-                self.prior.update_hyperparameters(hyperparameters)
-                self.likelihood.update(self.prior.hyperparameters)
-                self.marginal_density.update_hyperparameters()
-                logger.debug("    fvGP async hyperparameter update successful")
-                logger.debug("    Latest hyperparameters: {}", hyperparameters)
-            else:
-                logger.debug(
-                    "    The update was attempted but the new hyperparameters led to a \n \
-                    lower likelihood, so I kept the old ones")
-                logger.debug(f"Old likelihood: {-l_o} at {self.prior.hyperparameters}")
-                logger.debug(f"New likelihood: {-l_n} at {res}")
-        else:
-            logger.debug("    Async Hyper-parameter update not successful in fvGP. I am keeping the old ones.")
-            logger.debug("    hyperparameters: {}", self.prior.hyperparameters)
-
-        return self.prior.hyperparameters
-
-    ##################################################################################
-    def set_hyperparameters(self, hps):
-        """
-        Function to set hyperparameters.
-
-
-        Parameters
-        ----------
-        hps : np.ndarray
-            A 1-d numpy array of hyperparameters.
-        """
-        assert isinstance(hps, np.ndarray), "wrong format in hyperparameters"
-        assert np.ndim(hps) == 1, "wrong format in hyperparameters"
-        self.prior.update_hyperparameters(hps)
-        self.likelihood.update(self.prior.hyperparameters)
-        self.marginal_density.update_hyperparameters()
+        hps = self.trainer.update_hyperparameters(opt_obj)
+        self.set_hyperparameters(hps)
+        return hps
 
     ##################################################################################
     def get_hyperparameters(self):
@@ -777,8 +795,9 @@ class GP:
         ------
         hyperparameters : np.ndarray
         """
-
-        return self.prior.hyperparameters
+        warnings.warn('`get_hyperparameters()` is deprecated. Please use `hyperparameters`',
+                      DeprecationWarning, stacklevel=2)
+        return self.hyperparameters
 
     ##################################################################################
     def get_prior_pdf(self):
@@ -1413,6 +1432,35 @@ class GP:
         return (D ** 2 * tb) / (2. * n * b ** 2)
 
 
+    def initialize_gp2Scale_dask_client(self, gp2Scale, gp2Scale_dask_client):
+        if gp2Scale:
+            try:
+                from imate import logdet as imate_logdet
+            except:
+                raise Exception(
+                    "You have activated `gp2Scale`. You need to install imate"
+                    " manually for this to work.")
+            if gp2Scale_dask_client is None:
+                logger.debug("Creating my own local client.")
+                try: gp2Scale_dask_client = Client()
+                except: logger.debug("no client available")
+        return gp2Scale_dask_client
+
+    def __getstate__(self):
+        state = dict(
+            data=self.data,
+            prior=self.prior,
+            likelihood=self.likelihood,
+            marginal_density=self.marginal_density,
+            trainer=self.trainer,
+            posterior=self.posterior
+            )
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+
+
 ####################################################################################
 ####################################################################################
 ####################################################################################
