@@ -1,37 +1,25 @@
-from typing import Any
-
 import numpy as np
-from loguru import logger
 from .gp_lin_alg import *
-import warnings
 from scipy.sparse import issparse
 import scipy.sparse as sparse
+from loguru import logger
 
 
 class GPMarginalDensity:
     def __init__(self,
-                 data_obj,
-                 prior_obj,
-                 likelihood_obj,
-                 calc_inv=False,
-                 gp2Scale=False,
-                 gp2Scale_linalg_mode=None,
-                 compute_device='cpu',
-                 args=None):
+                 data,
+                 prior,
+                 likelihood,
+                 trainer,
+                 gp2Scale_linalg_mode=None):
 
-        self.data_obj = data_obj
-        self.prior_obj = prior_obj
-        self.likelihood_obj = likelihood_obj
-        self.calc_inv = calc_inv
-        self.gp2Scale = gp2Scale
-        self.compute_device = compute_device
+        self.data = data
+        self.prior = prior
+        self.likelihood = likelihood
         self.gp2Scale_linalg_mode = gp2Scale_linalg_mode
-        self.args = args
+        self.trainer = trainer
 
-        if self.gp2Scale and self.calc_inv:
-            self.calc_inv = False
-            warnings.warn("gp2Scale use forbids calc_inv=True; it has been set to False")
-        self.KVlinalg = KVlinalg(compute_device, self.args)
+        self.KVlinalg = KVlinalg(self.compute_device, data)
         K, V, m = self._get_KVm()
         if self.gp2Scale:
             mode = self._set_gp2Scale_mode(K)
@@ -41,16 +29,79 @@ class GPMarginalDensity:
         self.KVinvY = self._set_KVinvY(K, V, m, mode)
 
     ##################################################################
-    def update_data(self, append):
+    @property
+    def args(self):
+        return self.data.args
+
+    @property
+    def x_data(self):
+        return self.data.x_data
+
+    @property
+    def y_data(self):
+        return self.data.y_data
+
+    @property
+    def K(self):
+        return self.prior.K
+
+    @property
+    def m(self):
+        return self.prior.m
+
+    @property
+    def V(self):
+        return self.likelihood.V
+
+    @property
+    def ram_economy(self):
+        return self.data.ram_economy
+
+    @property
+    def gp2Scale(self):
+        return self.data.gp2Scale
+
+    @property
+    def compute_device(self):
+        return self.data.compute_device
+
+    @property
+    def calc_inv(self):
+        return self.data.calc_inv
+
+    @property
+    def hyperparameters(self):
+        return self.trainer.hyperparameters
+
+    ##################################################################
+    def compute_prior_covariance_matrix(self, x_data, hyperparameters):
+        return self.prior.compute_prior_covariance_matrix(x_data, hyperparameters)
+
+    def compute_mean(self, x_data, hyperparameters):
+        return self.prior.compute_mean(x_data, hyperparameters)
+
+    def dk_dh(self, x1, x2, hyperparameters, direction=None):
+        return self.prior.dk_dh(x1, x2, hyperparameters, direction=direction)
+
+    def dm_dh(self, x, hyperparameters):
+        return self.prior.dm_dh(x, hyperparameters)
+
+    def calculate_V(self, x, hyperparameters):
+        return self.likelihood.calculate_V(x, hyperparameters)
+
+    def calculate_V_grad(self, x, hyperparameters, direction=None):
+        return self.likelihood.calculate_V_grad(x, hyperparameters, direction=direction)
+
+    ##################################################################
+
+    def update_state_data(self, append):
         """Update the marginal PDF when the data has changed in data likelihood or prior objects"""
         logger.debug("Updating marginal density after new data was appended.")
         K, V, m = self._get_KVm()
-        if append:
-            self.KVinvY = self._update_KVinvY(K, V, m)
-        else:
-            self.KVinvY = self._set_KVinvY(K, V, m, self.KVlinalg.mode)
+        if append: self.KVinvY = self._update_KVinvY(K, V, m)
+        else: self.KVinvY = self._set_KVinvY(K, V, m, self.KVlinalg.mode)
 
-    def update_hyperparameters(self):
+    def update_state_hyperparameters(self):
         """Update the marginal PDF when if hyperparameters have changed"""
         logger.debug("Updating marginal density after new hyperparameters were appended.")
         K, V, m = self._get_KVm()
@@ -59,22 +110,22 @@ class GPMarginalDensity:
     ##################################################################
     def _update_KVinvY(self, K, V, m):
         """This updates KVinvY after new data was communicated"""
-        y_mean = self.data_obj.y_data - m
+        y_mean = self.y_data - m
         KV = self._addKV(K, V)
         self.KVlinalg.update_KV(KV)
         KVinvY = self.KVlinalg.solve(y_mean, x0=self.KVinvY).reshape(len(y_mean))
-        return KVinvY.reshape(len(y_mean))
+        return KVinvY.reshape(y_mean.shape)
 
     def _set_KVinvY(self, K, V, m, mode):
         """Set or reset KVinvY for new hyperparameters"""
-        y_mean = self.data_obj.y_data - m
+        y_mean = self.y_data - m
         KV = self._addKV(K, V)
         logger.debug("K+V computed")
         self.KVlinalg.set_KV(KV, mode)
         logger.debug("KVlinalg obj set")
         logger.debug("Solve in progress")
         KVinvY = self.KVlinalg.solve(y_mean).reshape(len(y_mean))
-        return KVinvY.reshape(len(y_mean))
+        return KVinvY.reshape(y_mean.shape)
 
     ##################################################################
     def compute_new_KVinvY(self, KV, m):
@@ -83,7 +134,7 @@ class GPMarginalDensity:
         This is only used by some posterior functions and in the log likelihood functions.
         This does not change the KV obj
         """
-        y_mean = self.data_obj.y_data - m
+        y_mean = self.y_data - m
         if self.gp2Scale:
             raise Exception("Can't compute a new KVinvY")
             #mode = self._set_gp2Scale_mode(KV)
@@ -92,8 +143,8 @@ class GPMarginalDensity:
             #    KVinvY = calculate_LU_solve(LU_factor, y_mean, args=self.args)
             #elif mode == "Chol":
             #    if issparse(KV): KV = KV.toarray()
-            #    Chol_factor = calculate_Chol_factor(KV, args=self.args)
-            #    KVinvY = calculate_Chol_solve(Chol_factor, y_mean, args=self.args)
+            #    Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
+            #    KVinvY = calculate_Chol_solve(Chol_factor, y_mean, compute_device=self.compute_device, args=self.args)
             #elif mode == "sparseCG":
             #    KVinvY = calculate_sparse_conj_grad(KV, y_mean, args=self.args)
             #elif mode == "sparseMINRES":
@@ -112,9 +163,9 @@ class GPMarginalDensity:
             #else:
             #    raise Exception("No mode in gp2Scale", mode)
         else:
-            Chol_factor = calculate_Chol_factor(KV, args=self.args)
-            KVinvY = calculate_Chol_solve(Chol_factor, y_mean, args=self.args)
-        return KVinvY.reshape(len(y_mean))
+            Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
+            KVinvY = calculate_Chol_solve(Chol_factor, y_mean, compute_device=self.compute_device, args=self.args)
+        return KVinvY.reshape(y_mean.shape)
 
     def compute_new_KVlogdet_KVinvY(self, K, V, m):
         """
@@ -122,7 +173,7 @@ class GPMarginalDensity:
         This is only used by the training (the log likelihood)
         """
         KV = self._addKV(K, V)
-        y_mean = self.data_obj.y_data - m
+        y_mean = self.y_data - m
         if self.gp2Scale:
             mode = self._set_gp2Scale_mode(KV)
             if mode == "sparseLU":
@@ -131,9 +182,9 @@ class GPMarginalDensity:
                 KVlogdet = calculate_LU_logdet(LU_factor, args=self.args)
             elif mode == "Chol":
                 if issparse(KV): KV = KV.toarray()
-                Chol_factor = calculate_Chol_factor(KV, args=self.args)
-                KVinvY = calculate_Chol_solve(Chol_factor, y_mean, args=self.args)
-                KVlogdet = calculate_Chol_logdet(Chol_factor, args=self.args)
+                Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
+                KVinvY = calculate_Chol_solve(Chol_factor, y_mean, compute_device=self.compute_device, args=self.args)
+                KVlogdet = calculate_Chol_logdet(Chol_factor, compute_device=self.compute_device, args=self.args)
             elif mode == "sparseCG":
                 KVinvY = calculate_sparse_conj_grad(KV, y_mean, args=self.args)
                 KVlogdet = calculate_random_logdet(KV, self.compute_device, args=self.args)
@@ -158,13 +209,13 @@ class GPMarginalDensity:
             else:
                 raise Exception("No mode in gp2Scale", mode)
         else:
-            Chol_factor = calculate_Chol_factor(KV, args=self.args)
-            KVinvY = calculate_Chol_solve(Chol_factor, y_mean, args=self.args)
-            KVlogdet = calculate_Chol_logdet(Chol_factor, args=self.args)
+            Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
+            KVinvY = calculate_Chol_solve(Chol_factor, y_mean, compute_device=self.compute_device, args=self.args)
+            KVlogdet = calculate_Chol_logdet(Chol_factor, compute_device=self.compute_device, args=self.args)
         return KVinvY.reshape(len(y_mean)), KVlogdet
 
     def _get_KVm(self):
-        return self.prior_obj.K, self.likelihood_obj.V, self.prior_obj.m
+        return self.K, self.V, self.m
 
     @staticmethod
     def _addKV(K, V):
@@ -199,13 +250,13 @@ class GPMarginalDensity:
             raise Exception("K+V not possible with the given formats")
 
     ##################################################################################
-    def _set_gp2Scale_mode(self, KV: object) -> str | list:
-        Ksparsity = float(KV.nnz) / float(len(self.data_obj.x_data) ** 2)
+    def _set_gp2Scale_mode(self, KV):
+        Ksparsity = float(KV.nnz) / float(len(self.x_data) ** 2)
         if self.gp2Scale_linalg_mode is not None:
             return self.gp2Scale_linalg_mode
-        elif len(self.data_obj.x_data) < 50001 and Ksparsity < 0.0001:
+        elif len(self.x_data) < 50001 and Ksparsity < 0.0001:
             mode = "sparseLU"
-        elif len(self.data_obj.x_data) < 2001 and Ksparsity >= 0.0001:
+        elif len(self.x_data) < 2001 and Ksparsity >= 0.0001:
             mode = "Chol"
         else:
             mode = "sparseMINRES"
@@ -232,18 +283,18 @@ class GPMarginalDensity:
             KVlogdet = self.KVlinalg.logdet()
         else:
             st = time.time()
-            K = self.prior_obj.compute_prior_covariance_matrix(self.data_obj.x_data, hyperparameters=hyperparameters)
+            K = self.compute_prior_covariance_matrix(self.x_data, hyperparameters)
             logger.debug("   Prior covariance matrix computed after {} seconds.", time.time() - st)
-            V = self.likelihood_obj.calculate_V(hyperparameters)
+            V = self.calculate_V(self.x_data, hyperparameters)
             logger.debug("   V computed after {} seconds.", time.time() - st)
-            m = self.prior_obj.compute_mean(self.data_obj.x_data, hyperparameters=hyperparameters)
+            m = self.compute_mean(self.x_data, hyperparameters)
             logger.debug("   Prior mean computed after {} seconds.", time.time() - st)
             KVinvY, KVlogdet = self.compute_new_KVlogdet_KVinvY(K, V, m)
             logger.debug("   KVinvY and logdet computed after {} seconds.", time.time() - st)
 
-        n = len(self.data_obj.y_data)
+        n = len(self.y_data)
 
-        return -(0.5 * ((self.data_obj.y_data - m).T @ KVinvY)) - (0.5 * KVlogdet) - (0.5 * n * np.log(2.0 * np.pi))
+        return -(0.5 * ((self.y_data - m).T @ KVinvY)) - (0.5 * KVlogdet) - (0.5 * n * np.log(2.0 * np.pi))
 
     ##################################################################################
     def neg_log_likelihood(self, hyperparameters=None):
@@ -280,28 +331,29 @@ class GPMarginalDensity:
 
         if hyperparameters is None:
             KVinvY = self.KVinvY
-            K = self.prior_obj.K
-            V = self.likelihood_obj.V
+            K = self.K
+            V = self.V
             KV = self._addKV(K, V)
+            hyperparameters = self.hyperparameters
         else:
-            K = self.prior_obj.compute_prior_covariance_matrix(self.data_obj.x_data, hyperparameters=hyperparameters)
-            V = self.likelihood_obj.calculate_V(hyperparameters)
-            m = self.prior_obj.compute_mean(self.data_obj.x_data, hyperparameters=hyperparameters)
+            K = self.compute_prior_covariance_matrix(self.x_data, hyperparameters)
+            V = self.calculate_V(self.x_data, hyperparameters)
+            m = self.compute_mean(self.x_data, hyperparameters)
             KV = self._addKV(K, V)
             KVinvY = self.compute_new_KVinvY(KV, m)
 
         b = KVinvY
         dK_dH = None
         a = None
-        if self.prior_obj.ram_economy is False:
+        if self.ram_economy is False:
             try:
-                noise_der = self.likelihood_obj.noise_function_grad(self.data_obj.x_data, hyperparameters)
+                noise_der = self.calculate_V_grad(self.x_data, hyperparameters)
                 assert np.ndim(noise_der) == 2 or np.ndim(noise_der) == 3
                 if np.ndim(noise_der) == 2:
-                    noise_der_V = np.zeros((len(hyperparameters), len(self.data_obj.x_data), len(self.data_obj.x_data)))
+                    noise_der_V = np.zeros((len(hyperparameters), len(self.x_data), len(self.x_data)))
                     for i in range(len(hyperparameters)): np.fill_diagonal(noise_der_V[i], noise_der[i])
                 else: noise_der_V = noise_der
-                dK_dH = self.prior_obj.dk_dh(self.data_obj.x_data, self.data_obj.x_data, hyperparameters) + noise_der_V
+                dK_dH = self.dk_dh(self.x_data, self.x_data, hyperparameters) + noise_der_V
             except Exception as e:
                 raise Exception(
                     "The gradient evaluation dK/dh + dNoise/dh was not successful. "
@@ -313,30 +365,30 @@ class GPMarginalDensity:
         bbT = np.outer(b, b.T)
         dL_dH = np.zeros((len(hyperparameters)))
         dL_dHm = np.zeros((len(hyperparameters)))
-        dm_dh = self.prior_obj.dm_dh(self.data_obj.x_data, hyperparameters)
+        dm_dh = self.dm_dh(self.x_data, hyperparameters)
 
         for i in range(len(hyperparameters)):
             dL_dHm[i] = -dm_dh[i].T @ b
-            if self.prior_obj.ram_economy is False:
+            if self.ram_economy is False:
                 matr = a[i]
             else:
                 try:
-                    noise_der = self.likelihood_obj.noise_function_grad(self.data_obj.x_data, i, hyperparameters)
+                    noise_der = self.calculate_V_grad(self.x_data, hyperparameters, direction=i)
                     assert np.ndim(noise_der) == 2 or np.ndim(noise_der) == 1
                     if np.ndim(noise_der) == 1:
-                        dK_dH = self.prior_obj.dk_dh(
-                            self.data_obj.x_data, self.data_obj.x_data, i, hyperparameters) + np.diag(noise_der)
+                        dK_dH = self.dk_dh(
+                            self.x_data, self.x_data, hyperparameters, direction = i) + np.diag(noise_der)
                     else:
-                        dK_dH = self.prior_obj.dk_dh(
-                            self.data_obj.x_data, self.data_obj.x_data, i, hyperparameters) + noise_der
-                except:
+                        dK_dH = self.dk_dh(
+                            self.x_data, self.x_data, hyperparameters, direction = i) + noise_der
+                except Exception as e:
                     raise Exception(
                         "The gradient evaluation dK/dh + dNoise/dh was not successful. "
                         "That normally means the combination of ram_economy and definition of "
-                        "the gradient function is wrong.")
+                        "the gradient function is wrong.",e)
                 matr = solve(KV, dK_dH, compute_device=self.compute_device)
             if dL_dHm[i] == 0.0:
-                if self.prior_obj.ram_economy is False:
+                if self.ram_economy is False:
                     mtrace = np.einsum('ij,ji->', bbT, dK_dH[i])
                 else:
                     mtrace = np.einsum('ij,ji->', bbT, dK_dH)
@@ -383,20 +435,41 @@ class GPMarginalDensity:
         analytical = -self.neg_log_likelihood_gradient(hyperparameters=thps)
         return grad, analytical
 
+    def __getstate__(self):
+        state = dict(
+            data=self.data,
+            prior=self.prior,
+            likelihood=self.likelihood,
+            calc_inv=self.calc_inv,
+            gp2Scale=self.gp2Scale,
+            compute_device=self.compute_device,
+            gp2Scale_linalg_mode=self.gp2Scale_linalg_mode,
+            KVlinalg=self.KVlinalg,
+            KVinvY=self.KVinvY
+        )
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+
 
 class KVlinalg:
-    def __init__(self, compute_device, args):
+    def __init__(self, compute_device, data):
         self.mode = None
+        self.data = data
         self.compute_device = compute_device
         self.KVinv = None
         self.KV = None
         self.Chol_factor = None
         self.LU_factor = None
         self.custom_obj = None
-        self.args = args
         self.allowed_modes = ["Chol", "CholInv", "Inv", "sparseMINRES", "sparseCG",
                               "sparseLU", "sparseMINRESpre", "sparseCGpre", "sparseSolve", "a set of callables"]
 
+    @property
+    def args(self):
+        return self.data.args
+
     def set_KV(self, KV, mode):
         self.mode = mode
         self.KVinv = None
@@ -406,11 +479,11 @@ class KVlinalg:
         assert self.mode is not None
         if self.mode == "Chol":
             if issparse(KV): KV = KV.toarray()
-            self.Chol_factor = calculate_Chol_factor(KV, args=self.args)
+            self.Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
         elif self.mode == "CholInv":
             if issparse(KV): KV = KV.toarray()
-            self.Chol_factor = calculate_Chol_factor(KV, args=self.args)
-            self.KVinv = calculate_inv_from_chol(self.Chol_factor, args=self.args)
+            self.Chol_factor = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
+            self.KVinv = calculate_inv_from_chol(self.Chol_factor, compute_device=self.compute_device, args=self.args)
         elif self.mode == "Inv":
             self.KV = KV
             self.KVinv = calculate_inv(KV, compute_device=self.compute_device, args=self.args)
@@ -435,18 +508,18 @@ class KVlinalg:
         if self.mode == "Chol":
             if issparse(KV): KV = KV.toarray()
             if len(KV) <= len(self.Chol_factor):
-                res = calculate_Chol_factor(KV, args=self.args)
+                res = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
             else:
-                res = update_Chol_factor(self.Chol_factor, KV, args=self.args)
+                res = update_Chol_factor(self.Chol_factor, KV, compute_device="cpu", args=self.args)
             self.Chol_factor = res
         elif self.mode == "CholInv":
             if issparse(KV): KV = KV.toarray()
             if len(KV) <= len(self.Chol_factor):
-                res = calculate_Chol_factor(KV, args=self.args)
+                res = calculate_Chol_factor(KV, compute_device=self.compute_device, args=self.args)
             else:
-                res = update_Chol_factor(self.Chol_factor, KV, args=self.args)
+                res = update_Chol_factor(self.Chol_factor, KV, compute_device="cpu", args=self.args)
             self.Chol_factor = res
-            self.KVinv = calculate_inv_from_chol(self.Chol_factor, args=self.args)
+            self.KVinv = calculate_inv_from_chol(self.Chol_factor, compute_device=self.compute_device, args=self.args)
         elif self.mode == "Inv":
             self.KV = KV
             if len(KV) <= len(self.KVinv):
@@ -472,10 +545,11 @@ class KVlinalg:
 
     def solve(self, b, x0=None):
         if self.mode == "Chol":
-            return calculate_Chol_solve(self.Chol_factor, b, args=self.args)
+            return calculate_Chol_solve(self.Chol_factor, b, compute_device=self.compute_device, args=self.args)
         elif self.mode == "CholInv":
-            return calculate_Chol_solve(self.Chol_factor, b, args=self.args)
+            return calculate_Chol_solve(self.Chol_factor, b, compute_device=self.compute_device, args=self.args)
         elif self.mode == "Inv":
+            #return matmul(self.KVinv, b, compute_device=self.compute_device) #is this really faster?
             return self.KVinv @ b
         elif self.mode == "sparseCG":
             return calculate_sparse_conj_grad(self.KV, b, x0=x0, args=self.args)
@@ -497,8 +571,8 @@ class KVlinalg:
             raise Exception("No Mode. Choose from: ", self.allowed_modes)
 
     def logdet(self):
-        if self.mode == "Chol": return calculate_Chol_logdet(self.Chol_factor, args=self.args)
-        elif self.mode == "CholInv": return calculate_Chol_logdet(self.Chol_factor, args=self.args)
+        if self.mode == "Chol": return calculate_Chol_logdet(self.Chol_factor, compute_device=self.compute_device, args=self.args)
+        elif self.mode == "CholInv": return calculate_Chol_logdet(self.Chol_factor, compute_device=self.compute_device, args=self.args)
         elif self.mode == "sparseLU": return calculate_LU_logdet(self.LU_factor, args=self.args)
         elif self.mode == "Inv": return calculate_logdet(self.KV, args=self.args)
         elif self.mode == "sparseCG": return calculate_random_logdet(self.KV, self.compute_device, args=self.args)
@@ -509,4 +583,19 @@ class KVlinalg:
         elif callable(self.mode[2]): return self.mode[2](self.custom_obj)
         else: raise Exception("No Mode. Choose from: ", self.allowed_modes)
 
+    def __getstate__(self):
+        state = dict(
+            mode=self.mode,
+            KVinv=self.KVinv,
+            KV=self.KV,
+            Chol_factor=self.Chol_factor,
+            LU_factor=self.LU_factor,
+            custom_obj=self.custom_obj,
+            allowed_modes=self.allowed_modes
+        )
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+
 
