@@ -2,6 +2,7 @@ import numpy as np
 from .kernels import *
 import dask.distributed as distributed
 import warnings
+warnings.simplefilter("once", UserWarning)
 import itertools
 from functools import partial
 import scipy.sparse as sparse
@@ -9,153 +10,189 @@ from scipy.sparse import block_array
 import time
 from loguru import logger
 from scipy.sparse import coo_matrix, vstack
+import inspect
 
 
 class GPprior:
     def __init__(self,
                  data,
+                 trainer,
                  kernel=None,
                  prior_mean_function=None,
                  kernel_grad=None,
                  prior_mean_function_grad=None,
-                 hyperparameters=None,
-                 ram_economy=False,
-                 compute_device='cpu',
-                 gp2Scale=False,
                  gp2Scale_dask_client=None,
                  gp2Scale_batch_size=10000,
-                 args=None
                  ):
 
-        assert callable(kernel) or kernel is None
-        assert callable(prior_mean_function) or prior_mean_function is None
-        assert isinstance(hyperparameters, np.ndarray)
-        assert np.ndim(hyperparameters) == 1
-
-        self.Euclidean = data.Euclidean
         self.kernel_function = kernel
         self.prior_mean_function = prior_mean_function
-        self.hyperparameters = hyperparameters
-        self.ram_economy = ram_economy
-        self.gp2Scale = gp2Scale
         self.client = gp2Scale_dask_client
         self.batch_size = gp2Scale_batch_size
         self.data = data
-        self.args = args
+        self.trainer = trainer
+
+        assert callable(kernel) or kernel is None
+        assert callable(prior_mean_function) or prior_mean_function is None
+        assert isinstance(self.hyperparameters, np.ndarray)
+        assert np.ndim(self.hyperparameters) == 1
 
         if not self.Euclidean and not callable(kernel):
             raise Exception(
                 "For GPs on non-Euclidean input spaces you need a user-defined kernel and initial hyperparameters.")
 
-        if gp2Scale:
+        if self.gp2Scale:
             if not callable(kernel):
                 warnings.warn("You have chosen to activate gp2Scale. A powerful tool!"
                               "But you have not supplied a kernel that is compactly supported."
                               "I will use an anisotropic Wendland kernel for now.",
                               stacklevel=2)
-                if compute_device == "cpu":
+                if self.compute_device == "cpu":
                     kernel = wendland_anisotropic_gp2Scale_cpu
-                elif compute_device == "gpu":
+                elif self.compute_device == "gpu":
                     kernel = wendland_anisotropic_gp2Scale_gpu
-            worker_info = list(self.client.scheduler_info()["workers"].keys())
-            if not worker_info: raise Exception("No workers available")
-            self.compute_workers = list(worker_info)
+            if self.client is not None:
+                worker_info = list(self.client.scheduler_info()["workers"].keys())
+                self.compute_workers = list(worker_info)
+            else: worker_info = False
+            if not worker_info: logger.debug("No workers available")
 
         # kernel
         if callable(kernel):
             self.kernel = kernel
+            k_n_params = len(inspect.signature(kernel).parameters)
         elif kernel is None:
             self.kernel = self._default_kernel
         else:
             raise Exception("No valid kernel function specified")
         self.d_kernel_dx = self._d_kernel_dx
         if callable(kernel_grad):
-            self.dk_dh = kernel_grad
+            self._dk_dh = kernel_grad
         else:
             if self.ram_economy is True:
-                self.dk_dh = self._kernel_derivative
+                self._dk_dh = self._kernel_derivative
             else:
-                self.dk_dh = self._kernel_gradient
+                self._dk_dh = self._kernel_gradient
 
-        # prior mean
+        # prior-mean
         if callable(prior_mean_function):
             self.mean_function = prior_mean_function
-            self.default_m = False
-        else:
-            self.mean_function = self._default_mean_function
-            self.default_m = True
+            m_n_params = len(inspect.signature(prior_mean_function).parameters)
+        else: self.mean_function = self._default_mean_function
 
-        if callable(prior_mean_function_grad):
-            self.dm_dh = prior_mean_function_grad
-        elif callable(prior_mean_function):
-            self.dm_dh = self._finitediff_dm_dh
-        else:
-            self.dm_dh = self._default_dm_dh
+        if callable(prior_mean_function_grad): self._dm_dh = prior_mean_function_grad
+        elif callable(prior_mean_function): self._dm_dh = self._finitediff_dm_dh
+        else: self._dm_dh = self._default_dm_dh
 
-        self.m, self.K = self._compute_prior(data.x_data)
+        self.m, self.K = self._compute_prior(data.x_data, self.hyperparameters)
         logger.debug("Prior successfully initialized.")
 
-    def augment_data(self, x_old, x_new):
-        self.m, self.K = self._update_prior(x_old, x_new)
+    ##############################################################
+    @property
+    def args(self):
+        return self.data.args
+
+    @property
+    def hyperparameters(self):
+        return self.trainer.hyperparameters
+
+    @property
+    def x_data(self):
+        return self.data.x_data
+
+    @property
+    def y_data(self):
+        return self.data.y_data
+
+    @property
+    def ram_economy(self):
+        return self.data.ram_economy
+
+    @property
+    def gp2Scale(self):
+        return self.data.gp2Scale
+
+    @property
+    def Euclidean(self):
+        return self.data.Euclidean
+
+    @property
+    def compute_device(self):
+        return self.data.compute_device
+
+    ################################################################
+    #START: FUNCTIONS THAT ALLOW INTERACTING WITH THE CLASS
+    def augment_state_data(self, x_old, x_new):
+        self.m, self.K = self._update_prior(x_old, x_new, self.hyperparameters)
         logger.debug("Prior mean and covariance updated after data augmentation.")
 
-    def update_data(self):
-        self.m, self.K = self._compute_prior(self.data.x_data)
+    def update_state_data(self):
+        self.m, self.K = self._compute_prior(self.x_data, self.hyperparameters)
         logger.debug("Prior mean and covariance updated after data change.")
 
-    def update_hyperparameters(self, hyperparameters):
-        self.hyperparameters = hyperparameters
-        self.m, self.K = self._compute_prior(self.data.x_data)
+    def update_state_hyperparameters(self):
+        self.m, self.K = self._compute_prior(self.x_data, self.hyperparameters)
         logger.debug("Prior mean and covariance updated after hyperparameter change.")
 
-    def _compute_prior(self, x_data):
-        m = self.compute_mean(x_data)
-        K = self.compute_prior_covariance_matrix(x_data)
+    def compute_prior_covariance_matrix(self, x, hyperparameters):
+        """computes the covariance matrix from the kernel"""
+        if self.gp2Scale:
+            K = self._compute_prior_covariance_gp2Scale(x, hyperparameters)
+        else:
+            K = self.compute_covariances(x, x, hyperparameters)
+        return K
+
+    def compute_covariances(self, x1, x2, hps):
+        return self.kernel(x1, x2, hps)
+
+    def compute_mean(self, x, hyperparameters):
+        """computes the covariance matrix from the kernel"""
+        m = self.mean_function(x, hyperparameters)
+        return m
+
+    def dk_dh(self, x1, x2, hyperparameters, direction=None):
+        #if direction is None: return self._dk_dh(x1, x2, hyperparameters)
+        #else: return self._dk_dh(x1, x2, direction, hyperparameters)
+        if self.ram_economy: return self._dk_dh(x1, x2, hyperparameters, direction)
+        else: return self._dk_dh(x1, x2, hyperparameters)
+
+    def dm_dh(self, x_data, hyperparameters):
+        return self._dm_dh(x_data, hyperparameters)
+    #END: FUNCTIONS THAT ALLOW INTERACTING WITH THE CLASS
+    #################################################################
+
+    def _compute_prior(self, x_data, hyperparameters):
+        m = self.compute_mean(x_data, hyperparameters)
+        K = self.compute_prior_covariance_matrix(x_data, hyperparameters)
         assert np.ndim(m) == 1
         assert np.ndim(K) == 2
         logger.debug("Prior mean and covariance matrix successfully computed.")
         return m, K
 
-    def _update_prior(self, x_old, x_new):
-        m = self._update_mean(x_new)
-        K = self._update_prior_covariance_matrix(x_old, x_new)
+    def _update_prior(self, x_old, x_new, hyperparameters):
+        m = self._update_mean(x_new, hyperparameters)
+        K = self._update_prior_covariance_matrix(x_old, x_new, hyperparameters)
         assert np.ndim(m) == 1
         assert np.ndim(K) == 2
         return m, K
 
-    def compute_prior_covariance_matrix(self, x, hyperparameters=None):
-        """computes the covariance matrix from the kernel"""
-        if hyperparameters is None: hyperparameters = self.hyperparameters
-        if self.gp2Scale:
-            K = self._compute_prior_covariance_gp2Scale(x, hyperparameters)
-        else:
-            K = self.kernel(x, x, hyperparameters)
-        return K
-
-    def _update_prior_covariance_matrix(self, x_old, x_new):
+    def _update_prior_covariance_matrix(self, x_old, x_new, hyperparameters):
         """This updated K based on new data"""
         if self.gp2Scale:
-            K = self._update_prior_covariance_gp2Scale(x_old, x_new, self.hyperparameters)
+            K = self._update_prior_covariance_gp2Scale(x_old, x_new, hyperparameters)
         else:
-            k = self.kernel(x_old, x_new, self.hyperparameters)
-            kk = self.kernel(x_new, x_new, self.hyperparameters)
+            k = self.compute_covariances(x_old, x_new, hyperparameters)
+            kk = self.compute_covariances(x_new, x_new, hyperparameters)
             K = np.block([
                 [self.K, k],
                 [k.T, kk]
             ])
         return K
 
-    def compute_mean(self, x_data, hyperparameters=None):
-        """computes the covariance matrix from the kernel"""
-        if hyperparameters is None: hyperparameters = self.hyperparameters
-        m = self.mean_function(x_data, hyperparameters)
-        return m
-
-    def _update_mean(self, x_new):
-        if self.default_m:
-            m = np.zeros((len(self.m) + len(x_new)))
-            m[:] = np.mean(self.data.y_data)
-        m = np.append(self.m, self.mean_function(x_new, self.hyperparameters))
+    def _update_mean(self, x_new, hyperparameters):
+        if np.ndim(self.m) == 1: m = np.append(self.m, self.compute_mean(x_new, hyperparameters))
+        elif np.ndim(self.m) == 2: m = np.vstack([self.m, self.compute_mean(x_new, hyperparameters)])
+        else: raise Exception("Prior mean in wrong format")
         return m
 
     @staticmethod
@@ -326,8 +363,8 @@ class GPprior:
         new_points = np.array(points1)
         epsilon = 1e-8
         new_points[:, direction] += epsilon
-        a = self.kernel(new_points, points2, hyperparameters)
-        b = self.kernel(points1, points2, hyperparameters)
+        a = self.compute_covariances(new_points, points2, hyperparameters)
+        b = self.compute_covariances(points1, points2, hyperparameters)
         derivative = (a - b) / epsilon
         return derivative
 
@@ -337,8 +374,7 @@ class GPprior:
             gradient[direction] = self._dkernel_dh(points1, points2, direction, hyperparameters)
         return gradient
 
-    def _kernel_derivative(self, points1, points2, direction, hyperparameters):
-        # gradient = np.empty((len(hyperparameters), len(points1),len(points2)))
+    def _kernel_derivative(self, points1, points2, hyperparameters, direction):
         derivative = self._dkernel_dh(points1, points2, direction, hyperparameters)
         return derivative
 
@@ -348,15 +384,20 @@ class GPprior:
         epsilon = 1e-8
         new_hyperparameters1[direction] += epsilon
         new_hyperparameters2[direction] -= epsilon
-        a = self.kernel(points1, points2, new_hyperparameters1)
-        b = self.kernel(points1, points2, new_hyperparameters2)
+        a = self.compute_covariances(points1, points2, new_hyperparameters1)
+        b = self.compute_covariances(points1, points2, new_hyperparameters2)
         derivative = (a - b) / (2.0 * epsilon)
         return derivative
 
     def _default_mean_function(self, x, hyperparameters):
         """evaluates the gp mean function at the data points """
-        mean = np.zeros((len(x)))
-        mean[:] = np.mean(self.data.y_data)
+        if np.ndim(self.y_data) == 1:
+            mean = np.zeros((len(x)))
+            mean[:] = np.mean(self.y_data)
+        elif np.ndim(self.y_data) == 2:
+            mean = np.zeros((len(x), self.y_data.shape[1]))
+            for i in range(mean.shape[1]): mean[:,i] = np.mean(self.y_data[:,i])
+        else: raise Exception("Wrong dim in default mean function")
         return mean
 
     def _finitediff_dm_dh(self, x, hps):
@@ -366,15 +407,36 @@ class GPprior:
             temp_hps1[i] = temp_hps1[i] + 1e-6
             temp_hps2 = np.array(hps)
             temp_hps2[i] = temp_hps2[i] - 1e-6
-            a = self.mean_function(x, temp_hps1)
-            b = self.mean_function(x, temp_hps2)
+            a = self.compute_mean(x, temp_hps1)
+            b = self.compute_mean(x, temp_hps2)
             gr[i] = (a - b) / 2e-6
         return gr
 
-    def _default_dm_dh(self, x, hps):
+    @staticmethod
+    def _default_dm_dh(x, hps):
         gr = np.zeros((len(hps), len(x)))
         return gr
 
+    def __getstate__(self):
+        state = dict(
+            kernel_function=self.kernel_function,
+            prior_mean_function=self.prior_mean_function,
+            batch_size=self.batch_size,
+            data=self.data,
+            trainer=self.trainer,
+            kernel=self.kernel,
+            d_kernel_dx=self.d_kernel_dx,
+            _dk_dh=self._dk_dh,
+            mean_function=self.mean_function,
+            _dm_dh=self._dm_dh,
+            m=self.m,
+            K=self.K,
+        )
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+
 
 ########################################################
 ########################################################
